#include "common.h"
#include "relays.h"
#include "spp_proc.h"
#include "spp.h"

#define CHANNEL_NB		RELNB

static void writeChannel(uint8_t channel, uint8_t cmd, uint8_t param)
{
	xprintf("wr: ch=%x, cmd=%x, p=%x\n",channel,cmd,param);
	
	if(channel > (CHANNEL_NB-1)) return;
	
	switch(cmd)
	{
		case (SPP_ID_CHAN_TOGGLE):		//just toggle
			relSet(channel, RELAY_TOGGLE);
			break;
		case(SPP_ID_CHAN_ON):	//just on
			relSet(channel, RELAY_ON);
			break;
		case(SPP_ID_CHAN_OFF):	//just off
			relSet(channel, RELAY_OFF);
			break;
		case(SPP_ID_CHAN_ON_ALT):	//alt (pulse) on
			relSet(channel, RELAY_PULSE_ON);
			break;
		case(SPP_ID_CHAN_OFF_ALT):	//alt (pulse) off
			relSet(channel, RELAY_PULSE_OFF);
			break;
		case(SPP_ID_CHAN_VALUE):	//arbitrary value normal (set relay if nonzero)
			if(param)
				relSet(channel, RELAY_ON);
			else
				relSet(channel, RELAY_OFF);
			break;
		case(SPP_ID_CHAN_VALUE_ALT):	//pulse on/off
			if(param)
				relSet(channel, RELAY_PULSE_ON);
			else
				relSet(channel, RELAY_PULSE_OFF);
			break;
		case(SPP_ID_TRANSITION):
			relSetPulseTime(channel, param);
			break;
	}
}

uint8_t sppRequestRxed(T_sppPacket* pPacket, uint8_t *reg)
{
	if((pPacket->dstAddr&0x0F) == SPP_BCAST_IN_CLASS) return 0;	//reject multicast packets
	if(pPacket->dstAddr == SPP_BCAST_ALL) return 0;			//reject broadcast packets
	
	if((pPacket->payload[0]&0xF0) == SPP_ID_REQ)
	{
		*reg = pPacket->payload[0]&0x0F;
		return 1;
	}
	return 0;
}


//0 - there was no request
//1 - there was a request but no valid response could be generated
//2 - there was a request and a valid response has been txed
int processRequest(T_sppPacket* pPacket)
{
	uint8_t reg;
	if(sppRequestRxed(pPacket,&reg))
	{
		pPacket->dstAddr = pPacket->srcAddr;
		pPacket->len = CHANNEL_NB + 1;
		pPacket->payload[0] = SPP_ID_RESP | reg;
		
		switch(reg)
		{
			case 0:	//get programmed power levels
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = relGetState(i);
				}
				sppTx(pPacket);
				return 2;
			}
			case 2:	//get current power levels
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = relGetState(i);
				}
				sppTx(pPacket);
				return 2;
			}
			case 3:	//get delays
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = relGetPulseTime(i);
				}
				sppTx(pPacket);
				return 2;
			}
			case 4:	//get pulse time left
			{
				for(int i=0;i<CHANNEL_NB;i++)
				{
					pPacket->payload[i+1] = relGetPulseTimeLeft(i);
				}
				sppTx(pPacket);
				return 2;
			}
		}
		
		return 1;
		
	}
	
	return 0;
}


int processGlobalCmd(T_sppPacket* pPacket)
{
	if(pPacket->len==0) return 0;
	uint8_t cmd = pPacket->payload[0];
	switch(cmd)
	{
		case SPP_ID_ALL_OFF:		//global off fast
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				relSet(i,RELAY_OFF);
			}
			return 2;
		
		case SPP_ID_ALL_OFF_ALT:	//global off alt
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				relSet(i,RELAY_PULSE_OFF);
			}
			return 2;
		
		case SPP_ID_ALL_ON:
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				relSet(i,RELAY_ON);
			}
			return 2;
		
		case SPP_ID_ALL_ON_ALT:
			for(uint8_t i=0;i<CHANNEL_NB;i++)
			{
				relSet(i,RELAY_PULSE_ON);
			}
			return 2;
	}
	return 0;
}


void exeMultichannel(uint8_t cmd, uint8_t param, uint16_t mask)
{
	//xprintf("multi: cmd=%x, p=%x, m=%x\n",cmd,param,mask);
	for(uint8_t i=0;i<CHANNEL_NB;i++)
	{
		if( ((uint16_t)1<<i) & mask )	//if channel is selected
		{
			writeChannel( i, cmd, param );
		}
	}
}


int processMultiChannel(T_sppPacket* pPacket)
{
	if(pPacket->payload[0]==SPP_ID_MULTICHANNEL)
	{
		if(pPacket->len < 4) return 0;	//a simple sanity check
		uint16_t chanMask = ((uint16_t)pPacket->payload[1]<<8) | pPacket->payload[2];
		uint8_t cmd = pPacket->payload[3] & 0xF0;	//ignore bits [3:0] because the channel info is in the mask
		uint8_t param = pPacket->payload[4];
		exeMultichannel(cmd,param,chanMask);
		return 1;
	}
	return 0;
}


void processChannelCommand(T_sppPacket* pPacket)
{
	if(pPacket->len == 0) return;
	
	uint8_t cmd = pPacket->payload[0] & 0xF0;
	uint8_t channel = pPacket->payload[0] & 0x0F;	//the first channel for this command
	
	switch(cmd)
	{
		case SPP_ID_CHAN_VALUE:
		case SPP_ID_CHAN_VALUE_ALT:
		case SPP_ID_POWER_DEFAULT:
		case SPP_ID_TRANSITION:
		{
			uint8_t dataLen = pPacket->len-1;
			for(uint8_t i=0;i<dataLen;i++)
			{
				uint8_t param = pPacket->payload[i+1];
				writeChannel(i+channel,cmd,param);
			}
			break;
		}
		default:
		{
			writeChannel(channel,cmd,0);
			break;
		}
	}
	
}

void processRxedPacket(T_sppPacket* pPacket)
{
	xprintf("spp rx: dst=%X src=%X len=%X cmd=%X\n",
		pPacket->dstAddr, pPacket->srcAddr, pPacket->len, pPacket->payload[0]);
	if(processRequest(pPacket) > 0) return;
	if(processMultiChannel(pPacket) > 0) return;
	if(processGlobalCmd(pPacket) > 0) return;
	processChannelCommand(pPacket);
}



