#include "spp.h"


enum{
	SPP_STATE_IDLE = 0,
	SPP_STATE_DSTADDR = 1,
	SPP_STATE_SRCADDR = 2,
	SPP_STATE_LEN = 3,
	SPP_STATE_PAYLOAD = 4,
	SPP_STATE_CRC = 5
};

static volatile uint8_t sppRxState = SPP_STATE_IDLE;


static volatile T_sppPacket rxPacket;	//rx buffer

#if SPP_DOUBLE_RX_BUFFER
	static volatile T_sppPacket readyPacket;	//the second buffer after complete tx
#endif


static void txb(uint8_t b);
static void rxIntEna(void);
static void rxIntDis(void);
static uint8_t txPacket_core(T_sppPacket* pPacket);
static uint8_t rxPacket_core(T_sppPacket* pPacket);
static void rxFsm(uint8_t rx);


static uint8_t crcByte(uint8_t crc, uint8_t data);

static volatile uint8_t rxReady = 0;


/*
	*** PLATFORM-DEPENDENT OBJECTS SHOULD BE INCLUDED BELOW ***
*/


#if (defined (__AVR_ATmega88__) || \
	defined (__AVR_ATmega8__)) 
		#include "spp_m8.h"
		
#elif (defined(AT91SAM7X128) || \
	defined (AT91SAM7X256) || \
	defined (AT91SAM7XC256) || \
	defined (AT91SAM7XC128))
		#include "spp_sam.h"

#elif (defined(STM32F10X_HD) || \
	defined(STM32F10X_MD) || \
	defined(STM32F10X_VL))
		#include "spp_stm32.h"
	
#else
	#warning Undefined target platform for SPPoB.
#endif


/*
	*** PLATFORM-INDEPENDENT CODE ***
*/


void sppInit(void)
{
	SPP_DIR_INIT;
	SPP_DIR_RX;
	SPP_LED_BUSY_INIT;
	SPP_LED_BUSY_OFF;
	sppRxState = SPP_STATE_IDLE;
	
	serialInit();
}


uint8_t sppIsIdle(void)
{
	if( sppRxState == SPP_STATE_IDLE ) return 1; else return 0;
}


#ifndef SPP_IDLE_TIMEOUT
	#define SPP_IDLE_TIMEOUT	100
	#warning SPP_IDLE_TIMEOUT not defined. Default value will be used.
#endif

void sppIdleTimeout(void)
{
	static uint16_t idleCntr;
	if(sppIsIdle())
	{
		idleCntr=0;
	}
	else
	{
		if(idleCntr < SPP_IDLE_TIMEOUT)
		{
			idleCntr++;
		}
		else
		{
			idleCntr=0;
			sppRxState = SPP_STATE_IDLE;
			SPP_LED_BUSY_OFF;
		}
	}
}


//tx with dispatch
static void txwd(uint8_t b)
{
	switch(b)
	{
		case 0x7E:
			txb(0x7D);
			txb(0x5E);
			return;
		
		case 0x7D:
			txb(0x7D);
			txb(0x5D);
			return;
		
		default:
			txb(b);
			return;
	}
}


//tx acknowledge, a wrapper to sppTx
//input: packet to be acked
//returns: 0 on error, nonzero if OK
uint8_t sppTxAck(T_sppPacket* pPacket)
{
	if((pPacket->dstAddr & 0xF0)==0xF0) return 0;	//don't ack broadcasts
	if((pPacket->dstAddr & 0x0F)==0x0F) return 0;	//don't ack broadcasts
	pPacket->dstAddr = pPacket->srcAddr;			//swap addresses = return to sender
	pPacket->len = 1;
	pPacket->payload[0] = SPP_ID_ACK;
	return(sppTx(pPacket));
}



//the receiver is in idle or ready state: 1=yes, 0=no
static uint8_t busIsFree(void)
{
	if(sppRxState == SPP_STATE_IDLE) return 1; else return 0;
}



//wait up to c.a. 100 ms for the bus to be idle
//returns 1 if the bus has been released before the timeout expired
//returns 0 if the timeout has expired and the bus is still busy
static uint8_t waitForFreeBus(void)
{
	uint16_t timeout = 100;
	while( timeout-- )
	{
		if(busIsFree())
		{
			return 1;
		}
		delay_ms(1);
	}
	return 0;
}


//returns zero if address is matched
static uint8_t checkDstAddress(uint8_t addr)
{
	//broadcast to all instancess of a class (addr[3:0]==0xF and addr[7:4] matches SPP_ADDRESS[7:4])
	if( ((addr&0x0F)==0x0F) && ((addr&0xF0)==(SPP_ADDRESS&0xF0)) ) return 0;
	
	//broadcast to all devices in the network
	if(addr==0xFF) return 0;
	
	//1:1 match
	if( addr==SPP_ADDRESS ) return 0;
	
	//mismatch
	return 1;
}


static uint8_t crcByte(uint8_t crc, uint8_t data)
{
	uint8_t i;
	crc = crc ^ data;
	for(i=0;i<8;i++)
	{
		if(crc & 0x01)
			crc = (crc >> 1) ^ 0x8C;
		else
			crc >>= 1;
	}
	return(crc);
}




static uint8_t txPacket_core(T_sppPacket* pPacket)
{
	SPP_LED_BUSY_ON;
	uint8_t waitRetriesLeft = 3;
	while((busIsFree()==0) && (waitRetriesLeft--))
	{
		waitForFreeBus();
		
		//other devices may attempt to tx just after the bus has become free
		//so we wait the unique period of time (directly dependent the on device address)
		//then we check the bus again
		delay_ms( SPP_ADDRESS * 2 );
	}
	
	//at this point we ignore the busy condition ->
	//...we assume that there was an error on the bus
	//(an unfinished packet transmission could cause that)
	
	//take over the bus
	txb(0x7E);
	
	uint8_t crc = 0;
	
	txwd(pPacket->dstAddr);
	crc = crcByte(crc,pPacket->dstAddr);
	
	txwd(SPP_ADDRESS);
	crc = crcByte(crc,SPP_ADDRESS);
	
	txwd(pPacket->len);
	crc = crcByte(crc,pPacket->len);
	
	for(uint8_t i=0;i<pPacket->len;i++)
	{
		txwd(pPacket->payload[i]);
		crc = crcByte(crc,pPacket->payload[i]);
	}
	
	txwd(crc);
	
	SPP_LED_BUSY_OFF;
	
	return 1;
}

//memcpy substitute
static void copyPacket(volatile T_sppPacket* dst, volatile T_sppPacket* src)
{
	dst->srcAddr = src->srcAddr;
	dst->dstAddr = src->dstAddr;
	dst->len = src->len;
	for(uint8_t i=0;i<src->len;i++)
		dst->payload[i] = src->payload[i];

}

static uint8_t rxPacket_core(T_sppPacket* pPacket)
{
	if(rxReady)
	{
		#if SPP_DOUBLE_RX_BUFFER
			rxIntDis();
			//copy from the second buffer
			copyPacket(pPacket,&readyPacket);
			
		#else
			//copy directly from the rx buffer
			copyPacket(pPacket,&rxPacket);
		#endif
		
		rxReady = 0;
		rxIntEna();
		
		SPP_LED_BUSY_OFF;
		
		return 1;
	}
	return 0;
}



static void rxFsm(uint8_t rx)
{
	static uint8_t dispatch;
	static uint8_t crc;
	static uint8_t payloadPtr;
	
	//address mismatch flag - track further packet bytes but ignore their content
	static uint8_t mismatch;
	
	switch(rx)
	{
		case 0x7E:	//rxed start byte
			sppRxState = SPP_STATE_DSTADDR;
			SPP_LED_BUSY_ON;
			dispatch = 0;
			mismatch = 0;
			return;	//exit!
		
		case 0x7D:
			dispatch = 1;
			return;	//exit!
		
		case 0x5E:
			if(dispatch) rx = 0x7E;
			break;
			
		case 0x5D:
			if(dispatch) rx = 0x7D;
			break;
	}
	
	dispatch = 0;
	
	switch(sppRxState)
	{
		case SPP_STATE_DSTADDR:
			#if SPP_SNIFF_MODE
				mismatch = 0;
			#else
				mismatch = checkDstAddress(rx);
			#endif
			rxPacket.dstAddr=rx;
			crc = crcByte(0,rx);	//start computing crc
			sppRxState = SPP_STATE_SRCADDR;
			
			break;
		
		case SPP_STATE_SRCADDR:
			rxPacket.srcAddr=rx;
			crc = crcByte(crc,rx);
			sppRxState = SPP_STATE_LEN;
			break;
		
		case SPP_STATE_LEN:
			if(rx > (SPP_PAYLOAD_LEN-1))
			{
				mismatch = 1;
			}
			rxPacket.len = rx;
			payloadPtr = 0;
			sppRxState = SPP_STATE_PAYLOAD;
			crc = crcByte(crc,rx);
			break;
		
		case SPP_STATE_PAYLOAD:
			if(payloadPtr>=rxPacket.len)	//end of payload?
			{
				#if !SPP_IGNORE_CRC
				if(crc != rx)
				{
					//Discard packet and go to IDLE state
					sppRxState = SPP_STATE_IDLE;
					SPP_LED_BUSY_OFF;
					return;
				}
				#endif
				
				if(mismatch==0)
				{
					//set rx state to "ready"
					sppRxState = SPP_STATE_IDLE;
					rxReady = 1;
					
					
					#if SPP_DOUBLE_RX_BUFFER
						//copy the rxedPacket contents to the second buffer
						copyPacket(&readyPacket,&rxPacket);
						//no need to disable further interrupts, the packet is now safe (at least for a while)
					#else
						//disable further RXready interrupts until rxedPacket is copied in main loop or a task
						rxIntDis();
					#endif
					
				}
				
				sppRxState = SPP_STATE_IDLE;
				SPP_LED_BUSY_OFF;
				
				return;
				
			}
			
			//is the packet for me?
			if(mismatch==0)
			{
				//write rxed bytes to the buffer and advance the pointer
				if(payloadPtr < (SPP_PAYLOAD_LEN-1))
					rxPacket.payload[payloadPtr++] = rx;
			}
			else
			{
				//keep track of the number of bytes rxed
				payloadPtr++;
			}
			
			crc = crcByte(crc,rx);
			break;
	}
	

}

